from typing import NamedTuple

import numpy as np


# 此枚举定义参考于vectorbt。多加了与保证金有关类别
class SizeTypeT(NamedTuple):
    # 空操作指令。通过此值比size全nan能减少代码执行
    NOP: int = 0
    # 下单数量和方向
    Amount: int = 1
    # 下单市值和方向
    Value: int = 2
    # 下单保证金和方向
    Margin: int = 3
    # 正数使用现金比例，负数卖出持仓比例
    Percent: int = 4
    # 目标数量和方向
    TargetAmount: int = 5
    # 目标市值和方向
    TargetValue: int = 6
    # 目标保证金和方向
    TargetMargin: int = 7
    # 目标市值百分比。size绝对值之和范围[0,1]
    TargetPercentValue: int = 8
    # 目标市值比例。size值可能为1.5:1:-1等
    TargetScaleValue: int = 9
    # 目标保证金百分比。size绝对值之和范围[0,1]
    TargetPercentMargin: int = 10
    # 目标保证金比例。size值可能为1.5:1:-1等
    TargetScaleMargin: int = 11


SizeType = SizeTypeT()

# 绩效统计。为减少内存，使用float32
performance_dt = np.dtype([
    ('date', np.int64),  # 日期时间
    ('asset', np.uint32),  # 资产ID
    ('amount', np.float32),  # 净持仓量。负数表示空头
    ('value', np.float32),  # 净持仓市值。负数表示空头
    ('cash', np.float32),  # 当前批量单执行后所剩余现金
    ('margin', np.float32),  # 占用保证金
    ('upnl', np.float32),  # 持仓盈亏
    ('cum_pnl', np.float32),  # 累计平仓盈亏（未扣除手续费）
    ('cum_commission', np.float32),  # 累计手续费
], align=True)

# 成交记录。为减少内存，使用float32
trade_dt = np.dtype([
    ('date', np.int64),  # 日期时间
    ('asset', np.uint32),  # 资产ID
    ('is_buy', np.bool_),  # 是否买入。
    ('is_open', np.bool_),  # 是否开平。开一定是开，平有可能含反手。反手也可以拆成两单
    ('fill_price', np.float32),  # 成交价
    ('qty', np.float32),  # 当前交易数量
    ('amount', np.float32),  # 持仓量和方向
    ('margin', np.float32),  # 保证金
    ('commission', np.float32),  # 手续费
    ('upnl', np.float32),  # 持仓盈亏
    ('pnl', np.float32),  # 平仓盈亏(未扣除手续费)
    ('cash_flow', np.float32),  # 现金流=平仓盈亏-保证金-手续费
    ('cash', np.float32),  # 现金
], align=True)

# 持仓记录。中间计算用字段与计算字段类型一致，而展示用字段减少内存
position_dt = np.dtype([
    ('asset', np.uint32),  # 资产ID
    # 用于相关下单数量计算
    ('mult', float),  # 合约乘数
    ('margin_ratio', float),  # 保证金率
    ('amount', float),  # 净持仓数量
    # 展示用字段
    ('value', np.float32),  # 市值
    ('open_value', np.float32),  # 开仓市值
    ('avg_price', np.float32),  # 平均价
    ('last_price', np.float32),  # 最新价
    ('margin', np.float32),  # 保证金
    ('upnl', np.float32),  # 持仓盈亏
    ('cum_pnl', np.float32),  # 累计平仓盈亏（未扣除手续费）
    ('cum_commission', np.float32),  # 累计手续费
], align=True)

# 外部下单指令，用于将用户的指令转成内部指令
order_outside_dt = np.dtype([
    ('date', np.int64),  # 日期时间
    ('size_type', int),  # size字段类型
    ('asset', int),  # 资产ID
    ('size', float),  # nan时表示此行不参与交易。可用于有持仓但不交易的资产更新最新价
    ('fill_price', float),  # 成交价
    ('last_price', float),  # 最新价
    ('date_diff', bool),  # 标记换日，会触发绩效更新
], align=True)

# 内部下单指令。用于将上层的目标持仓等信息转换成实际下单指令
order_inside_dt = np.dtype([
    ('asset', int),  # 资产ID
    ('is_buy', bool),  # 是否买入
    ('is_open', bool),  # 是否开仓
    ('fill_price', float),  # 成交价
    ('qty', float),  # 成交数量
], align=True)

# 成交统计。条目数一般等于资产数量
trades_stats_dt = np.dtype([
    ('asset', np.uint32),  # 资产ID
    ('start', np.int64),  # 第一条记录时间
    ('end', np.int64),  # 最后一条记录时间
    ('period', np.int64),  # 期
    ('total_count', np.uint32),  # 总条数
    ('buy_count', np.uint32),  # 买入条数
    ('sell_count', np.uint32),  # 卖出条数
    ('min_qty', np.float32),  # 最小交易量
    ('max_qty', np.float32),  # 最大交易量
    ('avg_qty', np.float32),  # 平均交易量
    ('avg_buy_qty', np.float32),  # 平均买入量
    ('avg_sell_qty', np.float32),  # 平均卖出量
    ('avg_buy_price', np.float32),  # 平均买入价
    ('avg_sell_price', np.float32),  # 平均卖出价
    ('total_commission', np.float32),  # 总手续费
    ('min_commission', np.float32),  # 最小手续费
    ('max_commission', np.float32),  # 最大手续费
    ('avg_commission', np.float32),  # 平均手续费
    ('avg_buy_commission', np.float32),  # 平均买入手续费
    ('avg_sell_commission', np.float32),  # 平均卖出手续费
], align=True)

# 交易每轮。由入场和出场组成
roundtrip_dt = np.dtype([
    ('asset', np.uint32),  # 资产ID
    ('is_long', np.bool_),  # 是否多头
    ('is_close', np.bool_),  # 是否已平
    ('qty', np.float32),  # 数量
    ('entry_date', np.int64),  # 入场时间
    ('entry_price', np.float32),  # 入场价
    ('entry_commission', np.float32),  # 入场手续费
    ('exit_date', np.int64),  # 出场时间
    ('exit_price', np.float32),  # 出场价
    ('exit_commission', np.float32),  # 出场手续费
    ('pnl', np.float32),  # 本轮平仓盈亏
    ('pnl_com', np.float32),  # 本轮平仓盈亏（已减手续费）
], align=True)

# 每轮统计
roundtrip_stats_dt = np.dtype([
    ('asset', np.uint32),  # 资产ID
    ('total_count', np.uint32),  # 总条数
    ('long_count', np.uint32),  # 多头条数
    ('short_count', np.uint32),  # 空头条数
    ('winning_count', np.uint32),  # 盈利条数
    ('losing_count', np.uint32),  # 亏损条数
    ('win_rate', np.float32),  # 胜率
], align=True)



from typing import List, Dict, Union

import numpy as np
import pandas as pd

from lightbt.enums import trades_stats_dt, roundtrip_stats_dt, roundtrip_dt
from lightbt.utils import groupby


def total_equity(perf: pd.DataFrame) -> pd.DataFrame:
    """总权益统计。

    Parameters
    ----------
    perf: pd.DataFrame
        输入为`bt.performances()`的输出

    Returns
    -------
    pd.DataFrame
        多加了总权益

    Examples
    --------
    >>> pd.options.plotting.backend = 'plotly'
    >>> perf = bt.performances()
    >>> equity = total_equity(perf)
    >>> equity.plot()

    """
    # 此处的amount返回的是净持仓数量
    agg = {'cash': 'last', 'value': 'sum', 'margin': 'sum', 'upnl': 'sum', 'cum_pnl': 'sum', 'cum_commission': 'sum', 'amount': 'sum'}
    p = perf.set_index(['date', 'asset']).groupby(by=['date']).agg(agg)
    # 总权益曲线。cash中已经包含了pnls和commissions
    p['equity'] = p['cash'] + p['margin'] + p['upnl']
    return p


def pnl_by_asset(perf, asset: Union[int, str], close: pd.DataFrame,
                 mapping_asset_int: Dict[str, int],
                 mapping_int_asset: Dict[int, str]) -> pd.DataFrame:
    """单资产的盈亏信息

    Parameters
    ----------
    perf: pd.DataFrame
        输入为`bt.performances()`的输出
    asset: int or str
        资产id
    close: pd.DataFrame
        行情
    mapping_asset_int: dict
        资产 字符串转数字
    mapping_int_asset: dict
        资产 数字转字符串

    Returns
    -------
    pd.DataFrame
        多加了盈亏曲线

    Examples
    --------
    >>> pd.options.plotting.backend = 'plotly'
    >>> pnls = pnls = pnl_by_asset(perf, 's_0000', df[['date', 'asset', 'CLOSE']], bt.mapping_asset_int, bt.mapping_int_asset)
    >>> pnls.plot()
    >>> pd.options.plotting.backend = 'matplotlib'
    >>> pnls[['PnL', 'CLOSE']].plot(secondary_y='CLOSE')

    """
    if pd.api.types.is_string_dtype(perf['asset']):
        if isinstance(asset, int):
            asset = mapping_int_asset.get(asset)
    elif pd.api.types.is_integer_dtype(perf['asset']):
        if isinstance(asset, str):
            asset = mapping_asset_int.get(asset)

    df1 = perf[perf['asset'] == asset]

    if close is None:
        df = df1
        agg = {'value': 'sum', 'margin': 'sum', 'upnl': 'sum', 'cum_pnl': 'sum', 'cum_commission': 'sum', 'amount': 'sum'}
    else:
        if pd.api.types.is_string_dtype(close['asset']):
            if isinstance(asset, int):
                close = close.copy()
                close['asset'] = close['asset'].map(mapping_asset_int)
        elif pd.api.types.is_integer_dtype(close['asset']):
            if isinstance(asset, str):
                close = close.copy()
                close['asset'] = close['asset'].map(mapping_int_asset)

        df2 = close[close['asset'] == asset]
        df = pd.merge(left=df1, right=df2, left_on=['date', 'asset'], right_on=['date', 'asset'])
        agg = {'value': 'sum', 'margin': 'sum', 'upnl': 'sum', 'cum_pnl': 'sum', 'cum_commission': 'sum', 'amount': 'sum', close.columns[-1]: 'last'}

    p = df.set_index(['date', 'asset']).groupby(by=['date']).agg(agg)
    # 盈亏曲线=持仓盈亏+累计盈亏+累计手续费
    p['PnL'] = p['upnl'] + p['cum_pnl'] - p['cum_commission']
    return p


def pnl_by_assets(perf: pd.DataFrame,
                  assets: Union[List[str], List[int]],
                  mapping_asset_int: Dict[str, int],
                  mapping_int_asset: Dict[int, str]) -> pd.DataFrame:
    """多个资产的盈亏曲线

    Parameters
    ----------
    perf: pd.DataFrame
        输入为`bt.performances()`的输出
    assets: list[int] or list[str]
        关注的资产列表
    mapping_asset_int: dict
        资产 字符串转数字
    mapping_int_asset: dict
        资产 数字转字符串

    Returns
    -------
    pd.DataFrame
        多资产盈亏曲线矩阵

    Examples
    --------
    >>> perf = bt.performances()
    >>> pnls = pnl_by_assets(perf, ['s_0000', 's_0100', 's_0300'], bt.mapping_asset_int, bt.mapping_int_asset)
    >>> pnls.plot()

    """
    if pd.api.types.is_string_dtype(perf['asset']):
        if isinstance(assets[0], int):
            assets = list(map(mapping_int_asset.get, assets))
    elif pd.api.types.is_integer_dtype(perf['asset']):
        if isinstance(assets[0], str):
            assets = list(map(mapping_asset_int.get, assets))

    # 单资产的盈亏曲线
    df = perf[perf['asset'].isin(assets)]
    df = df.set_index(['date', 'asset'])
    df['PnL'] = df['upnl'] + df['cum_pnl'] - df['cum_commission']
    return df['PnL'].unstack().fillna(method='ffill')


def trades_to_roundtrips(trades: np.ndarray, asset_count: int) -> np.ndarray:
    """多笔成交转为成对的交易轮

    Parameters
    ----------
    trades: np.ndarray
        全体成交记录
    asset_count: int
        总资产数。用于分配足够的空间用于返回

    Returns
    -------
    np.ndarray

    """
    trades = trades[trades['asset'].argsort(kind='stable')]
    groups = groupby(trades, by='asset', dtype=None)

    records = np.zeros(len(trades) // 2 + 1 + asset_count, dtype=roundtrip_dt)
    k = 0  # 目标位置
    for group in groups:
        # 每段开始位置
        flag = np.ones(shape=len(group) + 1, dtype=bool)
        flag[1:] = group['amount'] == 0.0
        flag[-1] = True
        idx = np.argwhere(flag).flatten()
        for i, j in zip(idx[:-1], idx[1:]):
            g = group[i:j]
            rec = records[k]

            is_open = g[g['is_open']]
            is_close = g[~g['is_open']]

            rec['asset'] = g['asset'][0]
            rec['is_long'] = g['is_buy'][0]
            rec['is_close'] = g['amount'][-1] == 0.0
            rec['qty'] = np.sum(is_open['qty'])
            rec['entry_date'] = g['date'][0]
            rec['entry_price'] = np.mean(is_open['fill_price'])
            rec['entry_commission'] = np.sum(is_open['commission'])
            rec['exit_date'] = g['date'][-1]
            if len(is_close) > 0:
                rec['exit_price'] = np.mean(is_close['fill_price'])
                rec['exit_commission'] = np.sum(is_close['commission'])
            rec['pnl'] = np.sum(g['pnl'])
            rec['pnl_com'] = rec['pnl'] - rec['entry_commission'] - rec['exit_commission']

            k += 1
    return records[:k]


def calc_roundtrips_stats(roundtrips: np.ndarray, asset_count: int) -> np.ndarray:
    """每轮交易统计

    Parameters
    ----------
    roundtrips: np.ndarray
        全体每轮交易
    asset_count: int
        总资产数。用于分配足够的空间用于返回

    Returns
    -------
    np.ndarray

    """
    roundtrips = roundtrips[roundtrips['asset'].argsort(kind='stable')]
    groups = groupby(roundtrips, by='asset', dtype=None)

    records = np.zeros(asset_count, dtype=roundtrip_stats_dt)
    for i, g in enumerate(groups):
        rec = records[i]

        is_long = g[g['is_long']]
        is_short = g[~g['is_long']]
        winning = g[g['pnl_com'] > 0.0]
        losing = g[g['pnl_com'] < 0.0]

        rec['asset'] = g['asset'][0]
        rec['total_count'] = len(g)
        rec['long_count'] = len(is_long)
        rec['short_count'] = len(is_short)
        rec['long_count'] = len(is_long)
        rec['short_count'] = len(is_short)
        rec['winning_count'] = len(winning)
        rec['losing_count'] = len(losing)
        rec['win_rate'] = rec['winning_count'] / rec['total_count']

    return records[:i + 1]


def calc_trades_stats(trades: np.ndarray, asset_count: int) -> np.ndarray:
    """成交统计

    Parameters
    ----------
    trades: np.ndarray
        全体交易记录
    asset_count: int
        总资产数。用于分配足够的空间用于返回

    Returns
    -------
    np.ndarray

    """
    trades = trades[trades['asset'].argsort(kind='stable')]  # stable一定要有，否则乱序
    groups = groupby(trades, by='asset', dtype=None)

    records = np.zeros(asset_count, dtype=trades_stats_dt)
    for i, g in enumerate(groups):
        rec = records[i]

        is_buy = g[g['is_buy']]
        is_sell = g[~g['is_buy']]

        rec['asset'] = g['asset'][0]
        rec['start'] = g['date'][0]
        rec['end'] = g['date'][-1]
        rec['period'] = rec['end'] - rec['start']
        rec['total_count'] = len(g)
        rec['buy_count'] = len(is_buy)
        rec['sell_count'] = len(is_sell)
        rec['min_qty'] = np.min(g['qty'])
        rec['max_qty'] = np.max(g['qty'])
        rec['avg_qty'] = np.mean(g['qty'])
        rec['total_commission'] = np.sum(g['commission'])
        rec['min_commission'] = np.min(g['commission'])
        rec['max_commission'] = np.max(g['commission'])
        rec['avg_commission'] = np.mean(g['commission'])

        if len(is_buy) > 0:
            rec['avg_buy_qty'] = np.mean(is_buy['qty'])
            rec['avg_buy_price'] = np.mean(is_buy['fill_price'])
            rec['avg_buy_commission'] = np.mean(is_buy['commission'])
        if len(is_sell) > 0:
            rec['avg_sell_qty'] = np.mean(is_sell['qty'])
            rec['avg_sell_price'] = np.mean(is_sell['fill_price'])
            rec['avg_sell_commission'] = np.mean(is_sell['commission'])

    return records[:i + 1]



import time
from typing import Optional

import numpy as np
import pandas as pd


def timer(func):
    def wrapper(*args, **kwargs):
        start_time = time.perf_counter()
        result = func(*args, **kwargs)
        end_time = time.perf_counter()
        print(f"{func.__name__} executed in {end_time - start_time} seconds")
        return result

    return wrapper


class Timer:
    def __init__(self):
        self.start_time = None

    def __enter__(self):
        self.start_time = time.perf_counter()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        end_time = time.perf_counter()
        print(f"code executed in {end_time - self.start_time} seconds")


def groupby(df: pd.DataFrame, by: str, dtype: Optional[np.dtype] = None) -> np.ndarray:
    """简版数据分组

    Parameters
    ----------
    df: pd.DataFrame
    by: str
    dtype: np.dtype
        指定类型能提速

    Returns
    -------
    np.ndarray
        迭代器

    Notes
    -----
    `df`一定要提前按`by`排序，否则结果是错的

    """
    if dtype is not None:
        # 控制同样的位置。否则record转dtype失败会导致效率低
        df = df[list(dtype.names)]

    if isinstance(df, pd.DataFrame):
        # recarray转np.ndarray
        arr = np.asarray(df.to_records(index=False), dtype=dtype)

        # 这里支持复合分组
        idx = df.groupby(by=by)['asset'].count().cumsum().to_numpy()
        idx = np.insert(idx, 0, 0)
    else:
        # 原数据是否需要复制？从代码上看没有复制之处
        arr = df  # .copy()

        dt = arr[by]
        flag = np.ones(shape=len(dt) + 1, dtype=bool)
        # 前后都为True
        flag[1:-1] = dt[:-1] != dt[1:]
        idx = np.argwhere(flag).flatten()

    for i, j in zip(idx[:-1], idx[1:]):
        # 由于标记的位置正好是每段的开始位置，所以j不需加1
        yield arr[i:j]



import time
from typing import List, Union

import numpy as np
import pandas as pd

from lightbt.stats import calc_trades_stats, calc_roundtrips_stats, trades_to_roundtrips


class LightBT:
    def __init__(self,
                 init_cash: float = 10000,
                 positions_precision: float = 1.0,
                 max_trades: int = 10000,
                 max_performances: int = 10000,
                 ) -> None:
        """初始化

        Parameters
        ----------
        positions_precision: float
            持仓精度
        max_trades: int
            记录成交的缓存大小。空间不足时将丢弃
        max_performances: int
            记录绩效的缓存大小。空间不足时将丢弃

        """
        from lightbt.portfolio import Portfolio

        self._init_cash = init_cash
        self._positions_precision = positions_precision
        self._max_trades = max_trades
        self._max_performances = max_performances

        self.pf = Portfolio(positions_precision=self._positions_precision,
                            max_trades=self._max_trades,
                            max_performances=self._max_performances)
        # 入金
        self.deposit(self._init_cash)

        # 底层没有资产名字符串，只有纯数字
        self.mapping_asset_int = {}
        self.mapping_int_asset = {}
        self.conf: pd.DataFrame = pd.DataFrame()

    def reset(self):
        """重置。不需要再次`setup`，只需要重新跑一次`run_`即可"""
        self.pf.reset()
        # 入初始资金
        self.deposit(self._init_cash)

    def setup(self, df: pd.DataFrame) -> None:
        """映射资产，配置合约乘数和保证金率

        同名的会进行替换

        Parameters
        ----------
        df: pd.DataFrame
            - asset
            - mult
            - margin_ratio
            - commission_ratio
                手续费率
            - commission_fn
                手续费计算函数

        """
        self.conf = pd.concat([self.conf, df])
        self.conf.drop_duplicates(subset='asset', keep='first', inplace=True)

        # 资产与底层持仓位置的映射
        conf = self.conf.reset_index(drop=True)
        self.mapping_int_asset = conf['asset'].to_dict()
        self.mapping_asset_int = {v: k for k, v in self.mapping_int_asset.items()}

        # 转成底层方便的格式
        asset = np.asarray(conf.index, dtype=int)
        mult = np.asarray(conf['mult'], dtype=float)
        margin_ratio = np.asarray(conf['margin_ratio'], dtype=float)
        commission_ratio = np.asarray(conf['commission_ratio'], dtype=float)
        commission_fn = np.asarray(conf['commission_fn'])

        # 调用底层的批量处理函数
        self.pf.setup(asset, mult, margin_ratio, commission_ratio)
        # 设置手续费函数
        for aid, fn in zip(asset, commission_fn):
            self.pf.set_commission_fn(aid, fn)

    def asset_str2int(self, strings: Union[List[str], str]) -> Union[List[int], int]:
        """资产转换。字符串转数字"""
        if isinstance(strings, list) and len(strings) == 1:
            strings = strings[0]
        if isinstance(strings, str):
            return self.mapping_asset_int.get(strings)

        return list(map(self.mapping_asset_int.get, strings))

    def asset_int2str(self, integers: Union[List[int], int]) -> Union[List[str], str]:
        """资产转换。数字转字符串"""
        if isinstance(integers, list) and len(integers) == 1:
            integers = integers[0]
        if isinstance(integers, int):
            return self.mapping_int_asset.get(integers)

        return list(map(self.mapping_int_asset.get, integers))

    def deposit(self, cash: float) -> float:
        """入金

        Parameters
        ----------
        cash: float

        Returns
        -------
        float

        Notes
        -----
        默认资金为0，所以交易前需要入金

        """
        return self.pf.deposit(cash)

    def withdraw(self, cash: float) -> float:
        """出金"""
        return self.pf.withdraw(cash)

    def positions(self, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """持仓记录"""
        records = self.pf.positions()
        if not readable:
            return records

        df = pd.DataFrame.from_records(records)
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def trades(self, return_all: bool, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """成交记录

        Parameters
        ----------
        return_all: bool
            返回所有记录或返回最近一批记录
        readable: bool
            返回可读格式

        Returns
        -------
        pd.DataFrame or np.ndarray

        """
        records = self.pf.trades(return_all)
        if not readable:
            return records

        df = pd.DataFrame.from_records(records)
        df['date'] = pd.to_datetime(df['date'])
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def performances(self, return_all: bool, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """绩效记录"""
        records = self.pf.performances(return_all)
        if not readable:
            return records

        df = pd.DataFrame.from_records(records)
        df['date'] = pd.to_datetime(df['date'])
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def trades_stats(self, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """成交统计"""
        trades = self.pf.trades(True)
        stats = calc_trades_stats(trades, len(self.mapping_int_asset))
        if not readable:
            return stats

        df = pd.DataFrame.from_records(stats)
        df['start'] = pd.to_datetime(df['start'])
        df['end'] = pd.to_datetime(df['end'])
        df['period'] = pd.to_timedelta(df['period'])
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def roundtrips(self, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """每轮交易记录"""
        trades = self.pf.trades(True)
        rounds = trades_to_roundtrips(trades, len(self.mapping_int_asset))
        if not readable:
            return rounds

        df = pd.DataFrame.from_records(rounds)
        df['entry_date'] = pd.to_datetime(df['entry_date'])
        df['exit_date'] = pd.to_datetime(df['exit_date'])
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def roundtrips_stats(self, readable: bool = True) -> Union[pd.DataFrame, np.ndarray]:
        """每轮交易统计"""
        rounds = self.roundtrips(False)
        stats = calc_roundtrips_stats(rounds, len(self.mapping_int_asset))
        if not readable:
            return stats
        df = pd.DataFrame.from_records(stats)
        df['asset'] = df['asset'].map(self.mapping_int_asset)
        return df

    def run_bar(self, arr) -> None:
        """同一时点，截面所有资产立即执行

        Parameters
        ----------
        arr
            - date
            - size_type
            - asset
            - size
                nan可用于只更新价格但不交易
            - fill_price
            - last_price
            - commission
            - date_diff

        """
        self.pf.run_bar2(arr)

    def run_bars(self, arrs) -> None:
        """多时点，循序分批执行

        Parameters
        ----------
        arrs
            - date
            - size_type
            - asset
            - size:
                nan可用于只更新价格但不交易
            - fill_price
            - last_price
            - commission
            - date_diff

        """
        for arr in arrs:
            self.pf.run_bar2(arr)


def warmup() -> float:
    """热身

    由于Numba JIT编译要占去不少时间，提前将大部分路径都跑一遍，之后调用就快了"""
    # import os
    # os.environ['NUMBA_DISABLE_JIT'] = '1'
    # pd.set_option('display.max_columns', None)
    # pd.set_option('display.width', 1000)

    from lightbt.enums import SizeType
    from lightbt.callbacks import commission_by_qty, commission_by_value
    from lightbt.enums import order_outside_dt
    from lightbt.signals import orders_daily
    from lightbt.utils import groupby

    symbols = [('510300', 1, 1, 0.001, commission_by_qty), ('IF2309', 300, 0.2, 0.0005, commission_by_value), ]
    config = pd.DataFrame.from_records(symbols, columns=['asset', 'mult', 'margin_ratio', 'commission_ratio', 'commission_fn'])

    df1 = pd.DataFrame({'asset': ['510300', 'IF2309'],
                        'size': [np.nan, -0.5],
                        'fill_price': [4.0, 4000.0],
                        'last_price': [4.0, 4000.0],
                        'date': '2023-08-01',
                        'size_type': SizeType.TargetPercentValue})

    df2 = pd.DataFrame({'asset': ['510300', 'IF2309'],
                        'size': [0.5, 0.5],
                        'fill_price': [4.0, 4000.0],
                        'last_price': [4.0, 4000.0],
                        'date': '2023-08-02',
                        'size_type': SizeType.TargetPercentValue})

    df = pd.concat([df1, df2])
    df['date'] = pd.to_datetime(df['date'])

    tic = time.perf_counter()

    bt = LightBT(init_cash=10000 * 50)
    bt.deposit(10000 * 20)
    bt.withdraw(10000 * 10)

    bt.setup(config)
    # 只能在setup后才能做map
    df['asset'] = df['asset'].map(bt.mapping_asset_int)

    bt.run_bars(groupby(orders_daily(df), by='date', dtype=order_outside_dt))

    bt.positions()
    bt.trades(return_all=True)
    bt.performances(return_all=True)
    bt.reset()

    toc = time.perf_counter()
    return toc - tic



import numpy as np
import pandas as pd
from numba import njit, float64, bool_


def state_to_action(state: pd.DataFrame):
    """将持仓状态转换成下单操作。

    1. 开头出现nan
    2. 中段出现nan

    Parameters
    ----------
    state: pd.DataFrame
        持仓状态宽表。值应当都是整数，浮点和nan都是不合法数据

    Returns
    -------
    pd.DataFrame
        下单操作

    """
    state = state.fillna(0)  # 防止nan计算有问题
    action = state.diff()
    action.iloc[0] = state.iloc[0]
    return action


def action_to_state(action: pd.DataFrame):
    """将操作转换成状态

    Parameters
    ----------
    action: pd.DataFrame
        下单操作宽表

    Examples
    --------
    s = pd.DataFrame({
    'a': [1, 1, 2, 0, -1, 0],
    'b': [np.nan, np.nan, 1, 0, 0, 0],
    })
    print(s)
    a = state_to_action(s)
    print(a)
    s = action_to_state(a)
    print(s)

    """
    action = action.fillna(0)
    return action.cumsum()


@njit(float64[:](bool_[:], bool_[:], bool_[:], bool_[:], bool_, bool_), fastmath=True, nogil=True, cache=True)
def signals_to_amount(is_long_entry: np.ndarray, is_long_exit: np.ndarray,
                      is_short_entry: np.ndarray, is_short_exit: np.ndarray,
                      accumulate: bool = False,
                      action: bool = False) -> np.ndarray:
    """将4路信号转换成持仓状态。适合按资产分组后的长表

    在`LongOnly`场景下，`is_short_entry`和`is_short_exit`输入数据值都为`False`即可

    Parameters
    ----------
    is_long_entry: np.ndarray
        是否多头入场
    is_long_exit: np.ndarray
        是否多头出场
    is_short_entry: np.ndarray
        是否空头入场
    is_short_exit: np.ndarray
        是否空头出场
    accumulate: bool
        遇到重复信号时是否累计
    action: bool
        返回持仓状态还是下单操作

    Returns
    -------
    np.ndarray
        持仓状态

    Examples
    --------
    ```python
    long_entry = np.array([True, True, False, False, False])
    long_exit = np.array([False, False, True, False, False])
    short_entry = np.array([False, False, True, False, False])
    short_exit = np.array([False, False, False, True, False])

    amount = signals_to_amount(long_entry, long_exit, short_entry, short_exit, accumulate=True, action=False)
    ```

    """
    _amount: float = 0.0  # 持仓状态
    _action: float = 0.0  # 下单方向
    out_amount = np.zeros(len(is_long_entry), dtype=float)
    out_action = np.zeros(len(is_long_entry), dtype=float)
    for i in range(len(is_long_entry)):
        if _amount == 0.0:
            # 多头信号优先级高于空头信号
            if is_long_entry[i]:
                _amount += 1.0
                _action = 1.0
            elif is_short_entry[i]:
                _amount -= 1.0
                _action = -1.0
        elif _amount > 0.0:
            if is_long_exit[i]:
                _amount -= 1.0
                _action = -1.0
            elif is_long_entry[i] and accumulate:
                _amount += 1.0
                _action = 1.0
        else:
            if is_short_exit[i]:
                _amount += 1.0
                _action = 1.0
            elif is_short_entry[i] and accumulate:
                _amount -= 1.0
                _action = -1.0
        out_amount[i] = _amount
        out_action[i] = _action

    if action:
        return out_action
    else:
        return out_amount


def orders_daily(df: pd.DataFrame) -> pd.DataFrame:
    """

    Parameters
    ----------
    df

    Returns
    -------
    pd.DataFrame
        1. 已经按时间进行了排序。sort_values
        2. 添加了日期标记，用于触发内部的绩效快照

    Notes
    -----
    有多处修改了数据，所以需要`copy`。`sort_values`隐含了`copy`

    """
    # 全体数据排序，并复制
    df = df.sort_values(by=['date'])

    # 按日期标记，每段的最后一条标记为True。一定要提前排序
    date_0 = df['date'].dt.date
    df['date_diff'] = date_0 != date_0.shift(-1, fill_value=0)

    return df


def orders_weekly(df: pd.DataFrame) -> pd.DataFrame:
    """

    Parameters
    ----------
    df

    Returns
    -------
    pd.DataFrame
        1. 已经按时间进行了排序。sort_values
        2. 添加了日期标记，用于触发内部的绩效快照

    Notes
    -----
    有多处修改了数据，所以需要`copy`。`sort_values`隐含了`copy`

    """
    # 全体数据排序，并复制
    df = df.sort_values(by=['date'])

    # 按日期标记，每段的最后一条标记为True。一定要提前排序
    date_0 = df['date'].dt.isocalendar().week
    # week的范围是1~52，所以可以fill_value写0 或>52的值
    df['date_diff'] = date_0 != date_0.shift(-1, fill_value=99)

    return df



import os

import numpy as np
from numba import objmode, types, njit, float64, typeof, bool_
from numba.experimental import jitclass

from lightbt.callbacks import commission_0

__TOL__: float = 1e-6


@njit(float64(float64, float64, float64), fastmath=True, nogil=True, cache=True)
def _value_with_mult(price: float, qty: float, mult: float) -> float:
    """计算市值"""
    if mult == 1.0:
        # 少算一次乘法，速度快一些
        return price * qty
    return price * qty * mult


@njit(float64(float64, float64, float64), fastmath=True, nogil=True, cache=True)
def _avg_with_mult(value: float, qty: float, mult: float) -> float:
    """计算均价"""
    if mult == 1.0:
        # 少算一次除法法，速度快一些
        return value / qty
    return value / qty / mult


@njit(float64(float64, float64), fastmath=True, nogil=True, cache=True)
def _net_cash_flow_with_margin(value: float, margin_ratio: float) -> float:
    """计算净现金流"""
    if margin_ratio == 1.0:
        # 少算一次乘法，速度快一些
        return value
    return value * margin_ratio


@njit(bool_(float64), fastmath=True, nogil=True)
def _is_zero(x: float) -> bool:
    """是否为0

    float，double分别遵循R32-24,R64-53的标准。
    所以float的精度误差在1e-6；double精度误差在1e-15
    """
    return (x <= __TOL__) and (x >= -__TOL__)


# 部分参考了SmartQuant部分代码，但又做了大量调整
class Position:
    Asset: int
    # TODO: 由于List中使用嵌套jitclass有问题，不得得将SubPosition简化成float
    # 多头数量
    LongQty: float
    # 空头数量
    ShortQty: float

    # !!! 本来应当将合约乘数，保证金率等信息存在Instrument对象中，但为了简化，安排在Position对象中
    # 合约乘数
    _mult: float
    # 保证金率
    _margin_ratio: float
    # 手续费率
    _commission_ratio: float

    # 最新价。用于计算Value/UPnL
    LastPrice: float
    # 持仓量。空头为负数
    Amount: float
    # 累计买入量
    QtyBought: float
    # 累计卖出量
    QtySold: float
    # 开仓均价
    AvgPx: float

    # 手续费
    _commission: float
    # 累计手续费
    _cum_commission: float
    # 盈亏
    _pnl: float
    # 累计盈亏
    _cum_pnl: float
    # 保证金
    _margin: float
    # 开仓市值
    _open_value: float
    # 市值流变化
    _value_flow: float
    # 净现金流
    _net_cash_flow: float
    # 现金流
    _cash_flow: float

    def __init__(self, asset: int) -> None:
        """初始化

        Parameters
        ----------
        asset: int
            预设的资产顺序ID
        """
        self.Asset = asset

        self._mult = 1.0
        self._margin_ratio = 1.0
        self._commission_ratio = 0.0
        self._commission_fn = commission_0

        self.reset()

    def reset(self):
        self.LongQty = 0.0
        self.ShortQty = 0.0
        self.LastPrice = 0.0
        self.Amount = 0.0
        self.QtyBought = 0.0
        self.QtySold = 0.0
        self.AvgPx = 0.0
        self._commission = 0.0
        self._cum_commission = 0.0
        self._margin = 0.0
        self._open_value = 0.0
        self._pnl = 0.0
        self._cum_pnl = 0.0
        self._value_flow = 0.0
        self._net_cash_flow = 0.0
        self._cash_flow = 0.0

    @property
    def is_long(self) -> bool:
        """是否多头"""
        return self.Amount >= 0

    @property
    def Qty(self) -> float:
        """持仓数量"""
        if self.Amount >= 0:
            return self.Amount
        else:
            return -self.Amount

    @property
    def Value(self) -> float:
        """持仓市值。受last_prce影响"""
        if self.Amount == 0:
            return 0.0
        return self.LastPrice * self.Amount * self._mult

    @property
    def OpenValue(self) -> float:
        """开仓市值"""
        if self.Amount < 0:
            return -self._open_value
        else:
            return self._open_value

    @property
    def Margin(self) -> float:
        """保证金占用"""
        return self._margin

    @property
    def UPnL(self) -> float:
        """持仓盈亏=持仓市值-开仓市值。受`last_prce`影响"""
        if self.Amount == 0:
            return 0.0
        return self.Value - self.OpenValue

    @property
    def PnL(self) -> float:
        """平仓盈亏(未减手续费)"""
        return self._pnl

    @property
    def CumPnL(self) -> float:
        """累计平仓盈亏(未减手续费)"""
        return self._cum_pnl

    @property
    def CumCommission(self) -> float:
        """累计手续费"""
        return self._cum_commission

    @property
    def CashFlow(self) -> float:
        """现金流=平仓盈亏-保证金-手续费"""
        return self._cash_flow

    @property
    def Equity(self) -> float:
        """持仓权益=保证金+浮动盈亏。受`last_price`影响"""
        return self.Margin + self.UPnL

    def calc_value(self, price: float, qty: float):
        """计算市值"""
        return _value_with_mult(price, qty, self._mult)

    def calc_commission(self, is_buy: bool, is_open: bool, value: float, qty: float) -> float:
        """计算手续费"""
        return self._commission_fn(is_buy, is_open, value, qty, self._commission_ratio)

    def openable(self, cash: float, value: float, commission: float) -> bool:
        """可开手数

        需考虑负价格情况
        1. 原油、天然气、电价都出现过负价格
        2. 套利合约的价差也可能是负数

        Parameters
        ----------
        cash: float
            分配的可用现金。注意：不是总现金。
        value: float
            报单市值
        commission: float
            手续费

        Returns
        -------
        bool
            是否成功

        """
        # TODO: 负手续费的情况下是如何处理？
        if cash < 0:
            return False

        if self._margin_ratio == 1.0:
            return (cash + commission) >= value
        else:
            return (cash + commission) >= value * self._margin_ratio

    def closable(self, is_long: bool) -> float:
        """可平手数

        Parameters
        ----------
        is_long: bool
            是否多头

        Returns
        -------
        float

        """
        if is_long:
            return self.LongQty
        else:
            return self.ShortQty

    def set_commission_fn(self, func=commission_0) -> None:
        self._commission_fn = func

    def setup(self, mult: float = 1.0, margin_ratio: float = 1.0, commission_ratio: float = 0.0) -> None:
        """配置资产信息

        Parameters
        ----------
        mult: float
            合约乘数
        margin_ratio: float
            保证金率
        commission_ratio: float
            手续费率

        """
        self._mult = mult
        self._margin_ratio = margin_ratio
        self._commission_ratio = commission_ratio

    def settlement(self) -> float:
        """结算。结算后可能产生现金变动

        1. 逆回购返利息
        2. 分红
        3. 手续费减免
        """
        return 0.0

    def update_last_price(self, last_price: float) -> None:
        """更新最新价。用于计算资金曲线

        Parameters
        ----------
        last_price: float

        """
        self.LastPrice = last_price

    def fill(self, is_buy: bool, is_open: bool, value: float, fill_price: float, qty: float, commission: float = 0.0) -> None:
        """通过成交回报，更新各字段

        Parameters
        ----------
        is_buy: bool
            是否买入
        is_open: bool
            是否开仓。反手需要标记成平仓
        value: float
            开仓市值
        fill_price: float
            成交价
        qty: float
            成交量
        commission: float
            手续费

        """
        self._net_cash_flow = 0.0
        self._cash_flow = 0.0

        # 计算开仓市值，平均价。返回改变的持仓市值
        self._calculate(is_open, value, fill_price, qty, commission, self._mult)

        if is_buy:
            self.QtyBought += qty
        else:
            self.QtySold += qty

        # 此处需要更新正确的子持仓对像。如股票买入时只能更新昨仓对象，而买入时只能更新今仓对象
        # !!!注意: is_open与is_long是有区别的
        if is_open:
            if is_buy:
                self.LongQty += qty
            else:
                self.ShortQty += qty
        else:
            if is_buy:
                self.ShortQty -= qty
            else:
                self.LongQty -= qty

        # 新持仓量需做部分计算后再更新
        self.Amount = self.QtyBought - self.QtySold
        # 净现金流
        self._net_cash_flow = -self._value_flow * self._margin_ratio
        # 现金流。考虑了盈亏和手续费
        self._cash_flow = self._pnl + self._net_cash_flow - commission
        # 保证金占用
        self._margin -= self._net_cash_flow
        # 更新最新价，用于计算盈亏
        self.LastPrice = fill_price
        # 累计盈亏
        self._cum_pnl += self._pnl
        # 累计手续费
        self._cum_commission += self._commission

        # 这几个值出现接近0时调整成0
        # 有了这个调整后，回测速度反而加快
        if _is_zero(self.Amount):
            self.Amount = 0.0
            self._margin = 0.0

    def _calculate_pnl(self, is_long: bool, avg_price: float, fill_price: float, qty: float, mult: float) -> float:
        """根据每笔成交计算盈亏。只有平仓才会调用此函数"""
        value: float = fill_price - avg_price if is_long else avg_price - fill_price
        return qty * value * mult

    def _calculate(self, is_open: bool, value: float, fill_price: float, qty: float, commission: float, mult: float) -> None:
        """更新开仓市值和平均价。需用到合约乘数。
        内部函数，不检查合法性。检查提前，有利于加快速度"""
        self._pnl = 0.0
        self._value_flow = 0.0
        self._commission = 0.0

        # 当前空仓
        if self.Amount == 0.0:
            self._value_flow = value  # _value_with_mult(fill_price, qty, mult)
            self._open_value = self._value_flow
            self.AvgPx = fill_price
            self._commission = commission
            return

        # 开仓。已经到这只能是加仓
        if is_open:
            # 开仓改变的市值流与外部计算结果一样
            self._value_flow = value  # _value_with_mult(fill_price, qty, mult)
            self._open_value += self._value_flow
            self.AvgPx = _avg_with_mult(self._open_value, self.Qty + qty, mult)
            self._commission = commission
            return

        # 平仓
        self._pnl = self._calculate_pnl(self.is_long, self.AvgPx, fill_price, qty, mult)

        if _is_zero(self.Qty - qty):
            # 清仓，市值流正好是之前持仓市值
            self._value_flow = -self._open_value
            self._open_value = 0.0
            self.AvgPx = 0.0
            self._commission = commission
            return
        elif self.Qty > qty:
            # 减仓
            self._value_flow = -_value_with_mult(self.AvgPx, qty, mult)
            self._open_value += self._value_flow
            self._commission = commission
            return

        # !!! 为简化外部操作。对于反手情况也支持，但is_open=False

        # 反手。平仓后开仓
        num: float = qty - self.Qty
        old_frozen_value: float = self._open_value
        new_frozen_value: float = _value_with_mult(fill_price, num, mult)
        self._open_value = new_frozen_value
        self.AvgPx = fill_price
        self._value_flow = new_frozen_value - old_frozen_value
        self._commission = commission
        return

    def to_record_position(self, rec: np.ndarray, detail: bool) -> np.ndarray:
        """持仓对象转持仓记录

        Parameters
        ----------
        rec: np.ndarray
        detail: bool

        Returns
        -------
        np.ndarray

        """
        rec['mult'] = self._mult
        rec['margin_ratio'] = self._margin_ratio
        rec['amount'] = self.Amount

        if detail:
            rec['upnl'] = self.UPnL
            rec['value'] = self.Value
            rec['open_value'] = self.OpenValue
            rec['avg_price'] = self.AvgPx
            rec['last_price'] = self.LastPrice
            rec['margin'] = self._margin
            rec['asset'] = self.Asset
            rec['cum_pnl'] = self._cum_pnl
            rec['cum_commission'] = self._cum_commission

        return rec

    def to_record_trade(self, rec: np.ndarray,
                        date: np.int64, is_buy: bool, is_open: bool, fill_price: float, qty: float, cash: float) -> np.ndarray:
        """订单对象转订单记录"""
        rec['asset'] = self.Asset
        rec['amount'] = self.Amount
        rec['margin'] = self._margin
        rec['commission'] = self._commission
        rec['upnl'] = self.UPnL  # 最新价会导致此项发生变化
        rec['pnl'] = self._pnl
        rec['cash_flow'] = self._cash_flow

        rec['date'] = date
        rec['is_buy'] = is_buy
        rec['is_open'] = is_open
        rec['fill_price'] = fill_price
        rec['qty'] = qty
        rec['cash'] = cash

        return rec

    def to_record_performance(self, rec: np.ndarray, date: np.int64, cash: float) -> np.ndarray:
        """转绩效"""
        rec['date'] = date
        rec['cash'] = cash
        rec['asset'] = self.Asset
        rec['amount'] = self.Amount
        rec['value'] = self.Value
        rec['margin'] = self._margin
        rec['upnl'] = self.UPnL
        # pnl与commission只记录了产生交易舜时的值，还需要累计值
        rec['cum_pnl'] = self._cum_pnl
        rec['cum_commission'] = self._cum_commission

        return rec

    def __str__(self) -> str:
        # f-string的实现方法比较特殊
        # https://github.com/numba/numba/issues/8969
        with objmode(string=types.unicode_type):  # declare that the "escaping" string variable is of unicode type.
            string = f'Position(Asset={self.Asset}, Value={self.Value}, OpenValue={self.OpenValue}, Margin={self.Margin}, UPnL={self.UPnL}, PnL={self.PnL}, Amount={self.Amount})'
        return string


# 这种写法是为了方便开关断点调试
if os.environ.get('NUMBA_DISABLE_JIT', '0') != '1':
    commission_fn_type = typeof(commission_0)

    Position = jitclass(Position,
                        [('_commission_fn', commission_fn_type)])



from numba import cfunc, float64, bool_


# 以下是手续费处理函数，用户也可以自己定义，通过setup进行设置

@cfunc(float64(bool_, bool_, float64, float64, float64))
def commission_0(is_buy: bool, is_open: bool, value: float, qty: float, commission_ratio: float) -> float:
    """0手续费"""
    return 0.0


@cfunc(float64(bool_, bool_, float64, float64, float64))
def commission_by_qty(is_buy: bool, is_open: bool, value: float, qty: float, commission_ratio: float) -> float:
    """按数量计算手续费"""
    return qty * commission_ratio


@cfunc(float64(bool_, bool_, float64, float64, float64))
def commission_by_value(is_buy: bool, is_open: bool, value: float, qty: float, commission_ratio: float) -> float:
    """按市值计算手续费"""
    return value * commission_ratio


@cfunc(float64(bool_, bool_, float64, float64, float64))
def commission_AStock(is_buy: bool, is_open: bool, value: float, qty: float, commission_ratio: float) -> float:
    """按市值计算手续费"""
    if is_open:
        commission = value * commission_ratio
    else:
        # 卖出平仓，多收千1的税
        commission = value * (commission_ratio + 0.001)

    if commission < 5.0:
        return 5.0
    return commission
